#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil

def clean():
    os.chdir(os.path.dirname(os.path.abspath(__file__)))

    for f in os.listdir('.'):
        if f.endswith('.rst') or f.endswith('.txt'):
            os.remove(f)
            print('removed %s' % f)

if __name__ == '__main__':
    clean()



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from fastr.utils.iohelpers import load_gpickle, load_json
FASTR_LOG_TYPE = 'none'  # Do not get info about fastr


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('infile', metavar='__fastr_extra_job_info__.json', help='result file to cat')
    parser.add_argument('path', help='path of the data to print')
    return parser


def fastr_cat(infile, path):
    import os
    from pprint import pprint

    path = path.strip('/').split('/')

    if os.path.isdir(infile):
        infile = os.path.join(infile, '__fastr_result__.pickle.gz')

        if not os.path.isfile(infile):
            print('ERROR: Input file does not exist!')
            return

    if infile.endswith('.pickle.gz'):
        try:
            job = load_gpickle(infile)
            infile = job.extrainfofile
        except IOError:
            print('ERROR: Could not open {} for reading'.format(infile))

    data = load_json(infile)

    for part in path:
        try:
            data = data[part]
        except KeyError:
            raise KeyError('{} not found in {}'.format(part, data.keys()))

    if isinstance(data, basestring):
        for line in data.split('\n'):
            print line
    else:
        pprint(data)


def main():
    """
    Print information from a job file
    """
    # No arguments were parsed yet, parse them now
    parser = get_parser()
    args = parser.parse_args()
    fastr_cat(args.infile, args.path)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

IS_TEST = True


def create_network():
    # Import the faster environment and set it up
    import fastr
    # Create a new network
    network = fastr.Network(id_='add_ints')

    # Create a source node in the network
    source1 = network.create_source(fastr.typelist['Int'], id_='source')

    # Create a new node in the network using toollist
    node = network.create_node(fastr.toollist['AddInt'], id_="add")

    # Create a link between the source output and an input of the addint node
    node.inputs['left_hand'] = source1.output

    # Create a constant node and link it to the value2 input of the addint node
    node.inputs['right_hand'] = [10]

    # Create a sink to save the data
    sink = network.create_sink(fastr.typelist['Int'], id_='sink')

    # Link the addint node to the sink
    sink.input = node.outputs['result']

    return network


def source_data(network):
    return {'source': [1, 'vfslist://example_data/add_ints/values']}


def sink_data(network):
    return {'sink': 'vfs://tmp/results/{}/result_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id)}


def main():
    network = create_network()

    # Execute
    network.draw_network()
    network.execute(source_data(network), sink_data(network))


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Null plugin for fastr
"""

import json
import os

import fastr
from fastr import typelist, vfs
from fastr.core.ioplugin import IOPlugin
from fastr.data import url
from fastr.datatypes import TypeGroup, URLType


class Reference(IOPlugin):
    """
    The Reference plugin is create to handle ``ref://`` type or URLs. These
    URLs are to make the sink just write a simple reference file to the data.
    The reference file contains the DataType and the value so the result can
    be reconstructed. It for files just leaves the data on disk by reference.
    This plugin is not useful for production, but is used for testing purposes.
    """
    scheme = 'ref'

    def __init__(self):
        # initialize the instance and register the scheme
        super(Reference, self).__init__()

    def push_sink_data(self, value, outurl, datatype=None):
        """
        Write out the sink data from the inpath to the outurl.

        :param str value: the path of the data to be pushed
        :param str outurl: the url to write the data to
        :param DataType datatype: the datatype of the data, used for determining
                                  the total contents of the transfer
        :return: None
        """
        fastr.log.info('Push sink called with: {}, {}'.format(value, outurl))
        self.setup()

        if datatype is None or issubclass(datatype, TypeGroup):
            previous_datatype = datatype.id if datatype is not None else None
            datatype = typelist.guess_type(value, options=datatype)
            fastr.log.info('Determined specific datatype as {} (based on {})'.format(datatype.id, previous_datatype))

        out_path = vfs.url_to_path(outurl, scheme='ref')
        out_dir = os.path.dirname(out_path)

        # Make sure the out directory exists or is created
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        if issubclass(datatype, URLType):
            # Set a target vfs in the output directory
            filename = url.basename(value)
            target_path = os.path.join(out_dir, filename)
            value_url = vfs.path_to_url(target_path)

            # Make sure data gets copied over correctly
            fastr.vfs.push_sink_data(value, value_url, datatype)

            # Set value to value_url now that the data has been pushed
            value = value_url

        result = {
            'value': value,
            'datatype': datatype.id
        }

        with open(out_path, 'w') as fh_out:
            json.dump(result, fh_out)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from fastr.datatypes import URLType
from fastr.utils.checksum import hashsum


class JsonFile(URLType):
    description = 'json file'
    extension = 'json'

    def _validate(self):
        try:
            with open(self.parsed_value) as fh_in:
                json.load(fh_in)
                return True

        except ValueError:
            return False

    def checksum(self):
        with open(self.parsed_value) as fh_in:
            data = json.load(fh_in)

        return hashsum([data])



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the tool manager class
"""

import fastr
from fastr import config
from fastr.core.objectmanager import ObjectManager
from fastr.core.tool import Tool


class ToolManager(ObjectManager):
    def populate(self):
        super(ToolManager, self).populate()
        fastr.ioplugins.create_ioplugin_tool()

    @property
    def object_class(self):
        return Tool

    def get_object_version(self, obj):
        return obj.command['version']

    def toolversions(self, tool):
        """
        Return a list of available versions for the tool

        :param tool: The tool to check the versions for. Can be either a `Tool` or a `str`.
        :return: List of version objects. Returns `None` when the given tool is not known.
        """
        return self.objectversions(tool)


if 'toollist' not in locals():
    #: The fastr toollist
    toollist = ToolManager(config.tools_path)


#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import json
import shellescape

FASTR_LOG_TYPE = 'none'


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('infile', nargs='?',
                        default=os.path.join(os.getcwd(), '__sink_data__.json'),
                        metavar='__sink_data__.json', help='result file to cat')
    parser.add_argument('--verbose', '-v', dest='verbose', action='store_true',
                        help='path of the data to print')
    parser.add_argument('--sinks',  dest='sinks', nargs="*",
                        help='list results for specified sinks')
    parser.add_argument('--nodes',  dest='nodes', nargs="*",
                        help='list results for specified nodes')
    parser.add_argument('--samples', dest='samples', nargs="*",
                        help='list result for all samples')
    return parser


def read_sink_data(infile):
    if os.path.isdir(infile):
        infile = os.path.join(infile, '__sink_data__.json')

    if os.path.isfile(infile) and os.access(infile, os.R_OK):
        with open(infile) as sink_data_file:
            data = json.load(sink_data_file)
    else:
        print('ERROR: could not read: {}'.format(infile))
        data = None

    return data


def switch_sample_sink(sink_data):
    # Switch samples and sinks
    samples_overview = dict()
    for sink, samples in sink_data.iteritems():
        for sample_id, sample in samples.iteritems():
            if sample_id not in samples_overview:
                samples_overview[sample_id] = dict()
            samples_overview[sample_id][sink] = sample
    return samples_overview


def print_sinks(sink_data, sink_ids, verbose):
    # Filter dict
    if sink_ids:
        filtered_dict = {sink_id: sink_data[sink_id] for sink_id in sink_ids}
    else:
        filtered_dict = sink_data

    # Print dict
    for sink, samples in sorted(filtered_dict.iteritems()):
        finished = 0
        cancelled = 0
        failed_samples_string = []
        for sample_id, sample in sorted(samples.iteritems()):
            if (sample['status'] == "JobState.cancelled") or (sample['status'] == "JobState.failed"):
                cancelled += 1
                for error_message in sample['errors']:
                    failed_samples_string.append('  {}: {}'.format(sample_id, error_message[2]))
            elif sample['status'] == "JobState.finished":
                finished += 1
        print('{} -- {} failed -- {} succeeded'.format(sink, cancelled, finished))
        if verbose:
            for error_message in failed_samples_string:
                print(error_message)


def print_samples(sink_data, sample_ids, verbose):
    samples_overview = switch_sample_sink(sink_data)

    # Filter dict
    if sample_ids:
        filtered_dict = { sample_id: samples_overview[sample_id] for sample_id in sample_ids }
    else:
        filtered_dict = samples_overview

    # Print_sinks(samples_overview)
    for sample, sinks in sorted(filtered_dict.iteritems()):
        finished = 0
        cancelled = 0
        failed_sink_string = []
        for _, sink in sorted(sinks.iteritems()):
            if sink['status'] == "JobState.cancelled" or sink['status'] == "JobState.failed":
                cancelled += 1
                for error_message in sink['errors']:
                    failed_sink_string.append(error_message[2])
            elif sink['status'] == "JobState.finished":
                finished += 1
        print('{} -- {} failed -- {} succeeded'.format(sample, cancelled, finished))
        if verbose:
            for error_message in set(failed_sink_string):
                print(error_message)


def print_job_result(job_file):
    from fastr.utils.iohelpers import load_gpickle, load_json
    from pprint import pprint

    job = load_gpickle(job_file)
    info_store = load_json(job.extrainfofile)

    print('\n\n===== JOB {} ====='.format(job.id))
    if hasattr(job, 'network_id'):
        print('Network: {}'.format(job.network_id))
    if hasattr(job, 'run_id'):
        print('Run: {}'.format(job.run_id))

    print('Node: {}'.format(job.node_id))
    print('Sample index: {}'.format(job.sample_index))
    print('Sample id: {}'.format(job.sample_id))
    print('Status: {}'.format(job.status))
    print('Timestamp: {}'.format(job.timestamp))
    print('Job file: {}'.format(job.logfile))

    command = info_store['process'].get('command', None)
    if command is not None:
        print('\nCommand:')
        print('List representation: {}'.format(command))
        printable_command = []
        for item in command:
            printable_command.append(shellescape.quote(item))
        print('String representation: {}'.format(' '.join(printable_command)))

    print('\nOutput data:')
    pprint(job.output_data)

    print('\nStatus history:')
    for timestamp, status in job.status_list:
        print('{}: {}'.format(timestamp, status))

    if 'errors' in info_store and isinstance(info_store['errors'], list):
        print('\n----- ERRORS -----')
        for job_error in info_store['errors']:
            print('- {e[0]}: {e[1]} ({e[2]}:{e[3]})'.format(e=job_error))
        print('------------------')

    if 'process' in info_store:
        if 'stdout' in info_store['process']:
            print('\n----- STDOUT -----')
            print(info_store['process']['stdout'])
            print('------------------')

        if 'stderr' in info_store['process']:
            print('\n----- STDERR -----')
            print(info_store['process']['stderr'])
            print('------------------')


def print_sample_sink(sink_data, dirname, sample_sink_tuples, verbose):
    for sink, sample in sample_sink_tuples:
        print('Tracing errors for sample {} from sink {}'.format(sample, sink))
        errors = sink_data[sink][sample]['errors']
        for error in errors:
            error_string = error[0].split(' ')[-1]
            node = error_string.split('___')
            result_pickle = os.path.join(dirname, node[1], sample, '__fastr_result__.pickle.gz')

            print_job_result(result_pickle)


def print_sample_node(sink_data, dirname, sample_node_tuples, verbose):
    for sample, node in sample_node_tuples:
        result_pickle = os.path.join(dirname, node, sample, '__fastr_result__.pickle.gz')
        print_job_result(result_pickle)


def main():
    """
    Trace samples/sinks from a run
    """
    # No arguments were parsed yet, parse them now
    parser = get_parser()
    args = parser.parse_args()

    sink_data = read_sink_data(args.infile)

    if not sink_data:
        exit(1)

    if args.sinks is not None and args.samples is not None:
        if len(args.sinks) == len(args.samples):
            sample_sink_tuples = zip(args.sinks, args.samples)
            print_sample_sink(sink_data, os.path.dirname(args.infile), sample_sink_tuples, args.verbose)
        else:
            print("ERROR nr of sinks does not match number of samples")
            exit(1)
    elif args.nodes is not None and args.samples is not None:
        if len(args.nodes) == len(args.samples):
            sample_node_tuples = zip(args.samples, args.nodes)
            print_sample_node(sink_data, os.path.dirname(args.infile), sample_node_tuples, args.verbose)
        else:
            print("ERROR nr of sinks does not match number of samples")
            exit(1)
    elif args.sinks is not None:
        print_sinks(sink_data, args.sinks, args.verbose)
    elif args.samples is not None:
        print_samples(sink_data, args.samples, args.verbose)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import json
# Try to import the faster cPickle if available and fall back when needed
try:
    import cPickle as pickle
except ImportError:
    import pickle
import os
import time

import fastr


def load_gpickle(path, retry_scheme=None):
    if retry_scheme is None:
        retry_scheme = (1, 3, 5)

    fastr.log.debug('Attempting to load {} with retry scheme {}'.format(path, retry_scheme))

    # Start with a non-delayed attempt
    retry_scheme = (0,) + tuple(retry_scheme)

    # Get the jobfile if given a job
    for attempt, delay in enumerate(retry_scheme):
        try:
            # Wait before trying to load the file
            fastr.log.debug('Attempt {} after {} seconds of delay'.format(attempt, delay))
            time.sleep(delay)

            try:
                with gzip.open(path, 'rb') as fh_in:
                    data = pickle.load(fh_in)
            except AttributeError as exception:
                # This is probably due to an Enum that is created on the fly
                # when a Tool is loaded
                fastr.log.warning('Cannot find attribute during unpickling,'
                                  ' this is probably an extension type that'
                                  'was created with an older fastr version, '
                                  'attempting to populate resources and try '
                                  'again. Original exception: {}'.format(exception))
                fastr.toollist.populate()
                with gzip.open(path, 'rb') as fh_in:
                    data = pickle.load(fh_in)

            return data
        except Exception as exception:
            pass

    raise exception


def save_gpickle(path, data):
    with gzip.open(path, 'wb') as fh_out:
        pickle.dump(data, fh_out)

        # Make sure the data gets flushed
        fh_out.flush()
        os.fsync(fh_out.fileno())


def save_json(path, data, indent=2):
    with open(path, 'w') as out_fh:
        json.dump(data, out_fh, indent=indent)


def load_json(path):
    with open(path, 'r') as in_fh:
        return json.load(in_fh)



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
os.chdir(os.path.dirname(os.path.abspath(__file__)))

os.system('python doc_clean.py')
print('python generate_modules.py ..' + os.path.sep + ' -d .' + os.path.sep + ' -s rst -f')
os.system('python generate_modules.py ..' + os.path.sep + ' -d .' + os.path.sep + ' -s rst -f')

if os.path.exists('fastr.core.test.rst'):
    os.remove('fastr.core.test.rst')
if os.path.exists('fastr.utils.rst'):
    os.remove('fastr.utils.rst')
if os.path.exists('setup.rst'):
    os.remove('setup.rst')
os.system('make html')
#os.system('make latexpdf')



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module manages datatypes. These datatypes are python classes generated
from the XML/JSON datatype files.
"""
import os
import xml.etree.ElementTree as ElementTree

import fastr
import fastr.datatypes
from fastr.core.pluginmanager import BasePluginManager
from fastr.core.version import Version
from fastr.data import url
from fastr.datatypes import BaseDataType, DataType, TypeGroup, EnumType, Deferred
import fastr.exceptions as exceptions
from fastr.utils.checksum import hashsum


class DataTypeManager(BasePluginManager):
    """
    The DataTypeManager hold a mapping of all DataTypes in the fast system and
    can create new DataTypes from files/data structures.
    """

    def __init__(self):
        """
        The DataTypeManager constructor will create a new DataTypeManager and
        populate it with all DataTypes it can find in the paths set in
        ``fastr.config.types_path``.

        :return: the created DataTypeManager
        """
        self.types_map = {}
        self._preferred_types = []
        super(DataTypeManager, self).__init__(fastr.config.types_path)

    @property
    def fullid(self):
        """
        The fullid of the datatype manager
        """
        return 'fastr://typelist'

    @property
    def plugin_class(self):
        """
        The PluginClass of the items of the BasePluginManager
        """
        return BaseDataType

    # Allow key to be a id string or DataType
    def __keytransform__(self, key):
        """
        Key transformation for this mapping. The key transformation allows
        indexing by both the DataType name as well as the DataType it self.

        :param key: The name of the requested datatype or the datatype itself
        :type key: fastr.datatypes.BaseDataType or str
        :return: The requested datatype
        """
        if self.isdatatype(key):
            if key.name in self.data and self.data[key.name] is key:
                return key.name
            else:
                raise exceptions.FastrDataTypeMismatchError('key DataType {} not {}'.format(key.name, type(self).__name__))
        else:
            return key

    def populate(self):
        """
        Populate Manager. After scanning for DataTypes, create the AnyType and set the preferred types
        """
        super(DataTypeManager, self).populate()

        # Add the any type
        self['AnyType'] = fastr.datatypes.AnyType
        self['AnyFile'] = fastr.datatypes.AnyFile
        self['Deferred'] = fastr.datatypes.Deferred

        # DataType specific extras
        self._preferred_types = [self.data[t] for t in fastr.config.preferred_types if t in self.data]

    @property
    def _instantiate(self):
        """
        Flag indicating that the plugin should NOT be instantiated prior to saving
        """
        return False

    def _print_key(self, key):
        if key.startswith('__') and key.endswith('__'):
            return None

        return key

    def has_type(self, name):
        """
        Check if the datatype with requested name exists

        :param str name: the name of the requested datatype
        :return: flag indicating if the datatype exists
        :rtype: bool
        """
        return name in self.data

    def poll_datatype(self, filename):
        """
        Poll an xml file to see if there is a definition of a datatype in it.

        :param str filename: path of the file to poll
        :return: tuple with (id, version, basetype) if a datatype is found or (None, None, None) if no datatype is found
        """
        if os.path.exists(filename):
            tree = ElementTree.parse(filename)
            root = tree.getroot()

            if root.tag not in ('type', 'typegroup'):
                message = 'Invalid root tag ({}) in file!'.format(root.tag)
                fastr.log.warning(message)
                return (None, None, None)

            id_ = root.get('id')
            version = Version(root.get('version'))

            return (id_, version, root.tag)
        else:
            message = '{} not a valid filename'.format(filename)
            fastr.log.warning(message)
            return (None, None, None)

    def get_type(self, name):
        """Read a type given a typename. This will scan all directories in
        types_path and attempt to load the newest version of the DataType.

        :param str name: Name of the datatype that should be imported in the system
        :return: the datatype with the requested name, or None if datatype is not found

        .. note:: If type is already in TypeManager it will not load anything and return the
                  already loaded version.
        """
        fastr.log.debug('Attemping to get datatype {}'.format(name))
        if name in self:
            return self[name]

        latest_version = Version('0.0')
        latest_filename = ''

        for path in fastr.config.types_path:
            filename = os.path.join(path, name + '.xml')
            if os.path.exists(filename):
                (pollname, version, _) = self.poll_datatype(filename)
                if pollname == name and version > latest_version:
                    latest_version = version
                    latest_filename = filename

        if latest_filename == '':
            message = 'Could not find type with name {}'.format(name)
            fastr.log.error(message)
            return None

        fastr.log.debug('Found {} (version {}) in {}'.format(name, latest_version, latest_filename))
        self._load_item(latest_filename)
        return self[name]

    def _store_item(self, name, value):
        """
        Store an item in the BaseManager, will ignore the item if the key is
        already present in the BaseManager.

        :param name: the key of the item to save
        :param value: the value of the item to save
        :return: None
        """
        super(DataTypeManager, self)._store_item(name, value)

        if value.id is not None:
            value.__module__ = 'fastr.datatypes'
            setattr(fastr.datatypes, value.id, value)

    def create_enumtype(self, type_id, options, name=None):
        """
        Create a python class based on an XML file. This function return a
        completely functional python class based on the contents of a DataType
        XML file.

        Such a class will be of type EnumType.

        :param str type_id: the id of the new class
        :param iterable options: an iterable of options, each option should be str
        :return: the newly created subclass of EnumType
        :raises FastrTypeError: if the options is not an iterable of str
        """
        if type_id in self:
            if self[type_id].options != set(options):
                raise exceptions.FastrDataTypeMismatchError('Conflicting definition of Enum {}!'
                                                            ' (options {} vs {})'.format(type_id,
                                                                                         self[type_id].options,
                                                                                         options))

            fastr.log.debug('Returning existing DataType {}!'.format(type_id))
            return self[type_id]

        attributes = {}

        try:
            if isinstance(options, str):
                options = (options,)

            attributes['_options'] = frozenset(options)
        except TypeError:
            message = 'options must be a iterable containing the valid options for the Enum, found options {}'.format(options)
            fastr.log.error(message)
            raise exceptions.FastrTypeError(message)

        if not all(isinstance(x, str) for x in attributes['_options']):
            message = 'all options for an Enum must be of type str, found options {}'.format(options)
            fastr.log.error(message)
            raise exceptions.FastrTypeError(message)

        attributes['parent'] = self
        attributes['_sourcepath'] = None
        attributes['_hash'] = hashsum([type_id, name, options])
        attributes['__module__'] = 'fastr.datatypes'

        supertypes = (EnumType,)

        fastr.log.debug('Creating EnumType {} from script'.format(type_id))
        out = type(type_id, supertypes, attributes)
        self[type_id] = out
        setattr(fastr.datatypes, out.id, out)
        return out

    def guess_type(self, value, exists=True, options=None, preferred=None):
        """
        Guess the DataType based on a value str.

        :param str value: the value to guess the type for
        :param options: The options that are allowed to be guessed from
        :type options: TypeGroup, DataType or tuple of DataTypes
        :param bool extists: Indicate the value exists (if file) and can be
                             checked for validity, if false skip validity check
        :param iterable preferred: An iterable of preferred types in case
                                   multiple types match.
        :return: The resulting DataType or None if no match was found
        :raises FastrTypeError: if the options argument is of the wrong type

        The function will first create a list of all candidate DataTypes.
        Subsequently, it will check for each candidate if the value would
        valid. If there are multiple matches, the config value for preferred
        types is consulted to break the ties. If non of the DataTypes are in
        the preferred types list, a somewhat random DataType will be picked
        as the most optimal result.
        """
        fastr.log.debug('Guesstype value: {}, options: {}'.format(value, options))
        if options is None:
            options = set([x for x in self.values() if issubclass(x, DataType)])
        elif issubclass(options, TypeGroup):
            options = set(options.members)
        elif issubclass(options, DataType):
            options = set((options,))
        elif isinstance(options, tuple):
            options = set(options)
        else:
            raise exceptions.FastrTypeError('Invalid type for options ({})'.format(options))

        fastr.log.debug('Guesstype options: {}'.format(options))
        candidates = set()

        scheme = None
        if url.isurl(value):
            scheme = url.get_url_scheme(value)
            if scheme != 'vfs':
                if scheme != 'val':
                    fastr.log.warning('Cannot determine DataType based on URL with scheme {}'.format(url.get_url_scheme(value)))
                    return None

        for option in options:
            if not issubclass(option, DataType):
                continue
            elif issubclass(option, Deferred):
                continue  # We never want to find a deferred (it is not a valid type for an instance)
            elif option.extension is None:
                candidates.add(option)
            elif isinstance(value, (str, unicode)) and value.endswith('.{}'.format(option.extension)):
                candidates.add(option)

        fastr.log.debug('Guesstype candidates: {}'.format(candidates))
        if len(candidates) == 0:
            fastr.log.debug('No valid combinations of options and candidates!')
            return None

        if len(candidates) != 1 and exists:
            # Test validity of value for each DataType
            final_candidates = []
            for candidate in candidates:
                temp = candidate(value)
                if temp.valid:
                    final_candidates.append(candidate)
        else:
            final_candidates = list(candidates)

        # Remove types in these order in case of mutliple matches, that means
        # the Int has precidence over the Boolean, Float and String in case
        # of a tie
        types_to_remove = ['String', 'Float', 'Boolean', 'Int', 'UnsignedInt']

        fastr.log.debug('Final candidates: {}'.format(final_candidates))

        if len(final_candidates) > 1:
            for type_to_remove in types_to_remove:
                if fastr.typelist[type_to_remove] in final_candidates:
                    final_candidates.remove(fastr.typelist[type_to_remove])
                if len(final_candidates) == 1:
                    break

        if len(final_candidates) == 0:
            return None
        elif len(final_candidates) == 1:
            fastr.log.debug('Matched a single type: {}'.format(final_candidates[0]))
            return final_candidates[0]
        else:
            fastr.log.debug('Multiple DataTypes match, trying to find a preferred match! Remaining candidates: {}'.format(final_candidates))

            # Get preferred types from argument list
            if preferred is not None:
                for type_ in preferred:
                    if type_ in final_candidates:
                        fastr.log.info('Found preferred match (from keyword): {}'.format(type_))
                        return type_

            # Get preferred information from the config
            for type_ in self._preferred_types:
                if type_ in final_candidates:
                    fastr.log.debug('Found preferred match: {}'.format(type_))
                    return type_

        fastr.log.debug('Mutliple matches, removing matches without extension')
        backup_candidate = final_candidates[0]
        final_candidates = [x for x in final_candidates if x.extension is not None]

        if len(final_candidates) == 1:
            return final_candidates[0]
        else:
            fastr.log.error('Multiple DataTypes match, but no preferred match, picking one at random! Remaining candidates: {}'.format(final_candidates))
            if len(final_candidates) > 1:
                return final_candidates[0]
            else:
                return backup_candidate

    def match_types(self, *args, **kwargs):
        """
        Find the match between a list of DataTypes/TypeGroups, see :ref:`resolve-datatype` for details

        :param args: A list of DataType/TypeGroup objects to match
        :param kwargs: A 'preferred' keyword argument can be used to indicate a list of DataTypes to prefer in case of ties (first has precedence over later in list)
        :return: The best DataType match, or None if no match is possible.
        :raises FastrTypeError: if not all args are subclasses of BaseDataType
        """
        options = self.match_types_any(*args)

        # Check if it is a preferred type
        if 'preferred' in kwargs and kwargs['preferred'] is not None:
            if not all([self.isdatatype(item) for item in kwargs['preferred']]):
                message = 'All preferred types must be DataTypes!'
                fastr.log.warning(message)

            preferred = kwargs['preferred']
        else:
            preferred = self._preferred_types

        if len(options) == 0:
            fastr.log.warning("No matching DataType available (args {})".format(args))
            return None
        elif len(options) == 1:
            # This is a perfect match, no preferences needed
            result = options.pop()

            return result
        else:
            # Get preferred information from the config
            for type_ in preferred:
                if type_ in options:
                    return type_

            fastr.log.debug("No preferred DataType matches, (options {}, preferred {})".format(options, preferred))
            return None

    def match_types_any(self, *args):
        """
        Find the match between a list of DataTypes/TypeGroups, see :ref:`resolve-datatype` for details

        :param args: A list of DataType/TypeGroup objects to match
        :return: A set with all DataTypes that match.
        :rtype: set
        :raises FastrTypeError: if not all args are subclasses of BaseDataType
        """
        # Find a single argument that is a list
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]

        # Remove typeless str
        args = tuple(arg for arg in args if arg != str)

        if not all([self.isdatatype(item) for item in args]):
            message = 'All arguments must be DataTypes! (Found {})'.format(args)
            fastr.log.error(message)
            raise exceptions.FastrTypeError(message)

        # In case there are no args
        if len(args) == 0:
            fastr.log.debug("No DataTypes given to match")
            return None

        # Create an initial options set (make sure to copy and not reference the set!)
        if issubclass(args[0], TypeGroup):
            options = set(args[0].members)
        else:
            options = set((args[0],))

        # Find intersection of all arguments
        for datatype in args[1:]:
            if isinstance(datatype, DataType):
                datatype = type(datatype)

            if issubclass(datatype, TypeGroup):
                options &= datatype.members
            else:
                options &= set((datatype,))

        return options

    @staticmethod
    def isdatatype(item):
        """
        Check if item is a valid datatype for the fastr system.

        :param item: item to check
        :return: flag indicating if the item is a fastr datatype
        :rtype: bool
        """
        return isinstance(item, (type, type)) and issubclass(item, BaseDataType)


if 'typelist' not in locals():
    typelist = DataTypeManager()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module holds the ExecutionPluginManager as well as the base-class
for all ExecutionPlugins.
"""

from abc import abstractmethod
from collections import deque
import functools
import sys
import threading
import traceback
import types

from enum import Enum

import fastr
from fastr import exceptions
from fastr.core.baseplugin import Plugin
from fastr.core.pluginmanager import PluginSubManager
from fastr.execution.job import Job, JobState, SourceJob
from fastr.utils import iohelpers


class JobAction(Enum):
    """
    Job actions that can be performed. This is used for checking
    if held jobs should be queued, held longer or be cancelled.
    """

    hold = 'hold'
    queue = 'queue'
    cancel = 'cancel'


class ExecutionPlugin(Plugin):
    """
    This class is the base for all Plugins to execute jobs somewhere. There are
    many methods already in place for taking care of stuff.

    There are fall-backs for certain features, but if a system already
    implements those it is usually preferred to skip the fall-back and
    let the external system handle it. There are a few flags to enable
    disable these features:

    * ``cls.SUPPORTS_CANCEL`` indicates that the plugin can cancel queued jobs
    * ``cls.SUPPORTS_HOLD_RELEASE`` indicates that the plugin can queue jobs in
      a hold state and can release them again (if not, the base plugin will
      create a hidden queue for held jobs). The plugin should respect the
      ``Job.status == JobState.hold`` when queueing jobs.
    * ``cls.SUPPORTS_DEPENDENCY`` indicate that the plugin can manage job
      dependencies, if not the base plugin job dependency system will be used
      and jobs with only be submitted when all dependencies are met.
    * ``cls.CANCELS_DEPENDENCIES`` indicates that if a job is cancelled it will
      automatically cancel all jobs depending on that job. If not the plugin
      traverse the dependency graph and kill each job manual.

      .. note:: If a plugin supports dependencies it is assumed that when a
                job gets cancelled, the depending job also get cancelled
                automatically!

    Most plugins should only need to redefine a few abstract methods:

    * :py:meth:`__init__ <fastr.execution.executionpluginmanager.ExecutionPlugin.__init__>`
      the constructor
    * :py:meth:`cleanup <fastr.execution.executionpluginmanager.ExecutionPlugin.__init__>`
      a clean up function that frees resources, closes connections, etc
    * :py:meth:`_queue_job <fastr.execution.executionpluginmanager.ExecutionPlugin._queue_job>`
      the method that queues the job for execution

    Optionally an extra job finished callback could be added:

    * :py:meth:`_job_finished <fastr.execution.executionpluginmanager.ExecutionPlugin._job_finished>`
      extra callback for when a job finishes

    If ``SUPPORTS_CANCEL`` is set to True, the plugin should also implement:

    * :py:meth:`_cancel_job <fastr.execution.executionpluginmanager.ExecutionPlugin._cancel_job>`
      cancels a previously queued job

    If ``SUPPORTS_HOLD_RELEASE`` is set to True, the plugin should also implement:

    * :py:meth:`_hold_job <fastr.execution.executionpluginmanager.ExecutionPlugin._hold_job>`
      hold_job a job that is currently held
    * :py:meth:`_release_job <fastr.execution.executionpluginmanager.ExecutionPlugin._release_job>`
      releases a job that is currently held

    If ``SUPPORTED_DEPENDENCY`` is set to True, the plugin should:

        * Make sure to use the ``Job.hold_jobs`` as a list of its dependencies

    Not all of the functions need to actually do anything for a plugin. There
    are examples of plugins that do not really need a ``cleanup``, but for
    safety you need to implement it. Just using a ``pass`` for the method could
    be fine in such a case.

    .. warning::

        When overwriting other functions, extreme care must be taken not to break
        the plugins working, as there is a lot of bookkeeping that can go wrong.
    """

    #: Indicates if the plugin can cancel queued jobs
    SUPPORTS_CANCEL = False

    #: Indicates if the plugin can queue jobs in a hold state and can release
    #: them again (if not, the base plugin will create a hidden queue for held
    #:  jobs)
    SUPPORTS_HOLD_RELEASE = False

    #: Indicate if the plugin can manage job dependencies, if not the base
    #: plugin job dependency system will be used and jobs with only be
    #: submitted when all dependencies are met.
    SUPPORTS_DEPENDENCY = False

    #: Indicates that when a job is cancelled the dependenceis
    CANCELS_DEPENDENCIES = False

    @abstractmethod
    def __init__(self, finished_callback=None, cancelled_callback=None, status_callback=None):
        """
        Setup the ExecutionPlugin

        :param finished_callback: the callback to call after a job finished
        :param cancelled_callback: the callback to call after a job cancelled
        :return: newly created ExecutionPlugin
        """
        super(ExecutionPlugin, self).__init__()

        # Pylint seems to be unable to figure out the .dict() member
        # pylint: disable=no-member

        self.job_status = {}
        self.job_dict = {}
        self.job_archive = {}
        self._finished_callback = finished_callback
        self._cancelled_callback = cancelled_callback
        self._status_callback = status_callback

        # Dict indicating the depending jobs for a certain jobs (who is waiting on the key job id)
        self.held_queue = {}
        self.held_queue_lock = threading.Lock()

        # A list for the source jobs that should be held
        self.source_job_queue = deque()
        self.number_of_source_jobs_queued = 0
        self.source_job_limit = fastr.config.source_job_limit

        # Flag indicating the plugin is accepting new jobs queued
        self.accepting = True

    def __enter__(self):
        return self

    def __exit__(self, type_, value, tb):
        self.cleanup()

    def __del__(self):
        """
        Cleanup if the variable was deleted on purpose
        """
        fastr.log.debug('Calling cleanup')
        self.cleanup()

    @abstractmethod
    def cleanup(self):
        """
        Method to call to clean up the ExecutionPlugin. This can be to clear
        temporary data, close connections, etc.

        :param force: force cleanup (e.g. kill instead of join a process)
        """
        # Stop accepting new jobs (close the queue)
        self.accepting = False

        # Cancel all queued jobs
        while len(self.job_dict) > 0:
            job_id, job = self.job_dict.popitem()
            fastr.log.debug('Cleanup cancelling {}'.format(job_id))
            self.cancel_job(job)

    def register_job(self, job):
        self.job_dict[job.id] = job
        self.job_status[job.id] = job.status

        for hold_id in job.hold_jobs:
            # Add job reference to held queue to receive signal when the
            # required jobs are finished/failed. Do not subscribe for jobs that
            # are already finished.
            if hold_id in self.job_status and self.job_status[hold_id].done:
                continue

            with self.held_queue_lock:
                if hold_id not in self.held_queue:
                    self.held_queue[hold_id] = []

                # append to held_queue, because of the managed dict, we need to replace the value (not update)
                self.held_queue[hold_id].append(job.id)

    def _dispatch_job(self, job):
        # Check the job requirements for dependencies and take appropriate
        # action while keeping in mind the plugin capabilities.
        action = self.check_job_requirements(job.id)
        if action == JobAction.cancel:
            self.cancel_job(job)
        elif self.SUPPORTS_DEPENDENCY or action == JobAction.queue:
            self.job_status[job.id] = job.status = JobState.queued
            self._queue_job(job)
        else:  # Thus action has to be JobAction.hold:
            fastr.log.debug('Holding {} until dependencies are met'.format(job.id))
            self.job_status[job.id] = job.status = JobState.hold

            # If there is support for hold release the plugin should respect
            # the job.status being JobState.hold and it is safe to queue in
            # a held state already.
            if self.SUPPORTS_HOLD_RELEASE:
                self._queue_job(job)

    def queue_job(self, job):
        """
        Add a job to the execution queue

        :param Job job: job to add
        """
        if not self.accepting:
            return

        if isinstance(job, list):
            for j in job:
                self.queue_job(j)
            return

        # Register the job for tracking
        self.register_job(job)

        # Save the job (and initial result) to file before serializing
        iohelpers.save_gpickle(job.commandfile, job)

        # If the job has been preset to done, immediately send for the callback
        if job.status in [JobState.execution_done, JobState.execution_failed]:
            self.job_finished(job)
            return

        # Check if it is a source job and limit if needed
        if isinstance(job, SourceJob):
            if self.number_of_source_jobs_queued < self.source_job_limit:
                fastr.log.debug('Claiming source job slot {} -> {}'.format(self.number_of_source_jobs_queued,
                                                                           self.number_of_source_jobs_queued + 1))
                self.number_of_source_jobs_queued += 1
            else:
                # Keep the job in a temporary queue
                fastr.log.debug('Putting sour job {} in source queue due to limitations!'.format(job.id))
                self.source_job_queue.append(job)

                if self.SUPPORTS_HOLD_RELEASE:
                    # Submit the job in a hold state
                    fastr.log.debug('Queue SourceJob {} as hold'.format(job.status))
                    job.status = JobState.hold
                    self._queue_job(job)

                # Stop processing this job
                return

        # Dispatch the job to the plugin
        self._dispatch_job(job)

    def cancel_job(self, job):
        """
        Cancel a job previously queued

        :param job: job to cancel
        """
        if not isinstance(job, Job):
            try:
                job = self.job_dict[job]
            except KeyError:
                fastr.log.warning('Job {} is no longer under processing, cannot cancel!'.format(job))
                return

        # Check if job is not already being cancelled
        if job.status == JobState.cancelled:
            return

        if job.status == JobState.hold and not self.SUPPORTS_CANCEL and self.SUPPORTS_HOLD_RELEASE:
            # This is a corner case where the job is scheduled and held, but
            # somehow cannot be cancelled. We need to release the Job and let
            # it crash itself.
            self.release_job(job)

        if not self.SUPPORTS_CANCEL and job.status not in [JobState.created, JobState.hold]:
            fastr.log.warning('Job is already queued or running and cannot be cancelled anymore!')
            return

        fastr.log.debug('Cancelling {}'.format(job.id))

        # Cancel job
        old_status = job.status
        job.status = self.job_status[job.id] = JobState.cancelled

        # If supported, send out the actual cancellation command
        # If the job status is created, it means it never got properly
        # accepted by the plugin and thus only needs the callback, but
        # not the actual cancellation
        if self.SUPPORTS_CANCEL and old_status != JobState.created:
            self._cancel_job(job)

        fastr.log.debug('Removing {} from jobdict'.format(job.id))
        self.job_archive[job.id] = job

        # Initialize provenance and write the final state of the job to
        # the logfile
        job.provenance.init_provenance(job)
        job.write()

        try:
            del self.job_dict[job.id]
        except KeyError:
            pass

        # Calling callback for cancelling
        fastr.log.debug('Calling cancelled for {}'.format(job.id))
        if self._cancelled_callback is not None:
            self._cancelled_callback(job)

        # Cancel all children (some systems might have automatically cancelled them)
        fastr.log.debug('Cancelling children for {}'.format(job.id))

        if not self.CANCELS_DEPENDENCIES:
            if job.id in self.held_queue:
                fastr.log.debug('Found children....')
                held_queue = self.held_queue[job.id]
                for dependent_job in held_queue:
                    fastr.log.debug('Checking sub {}'.format(dependent_job))
                    if dependent_job in self.job_dict and dependent_job in self.job_status and not self.job_status[dependent_job].done:
                        fastr.log.debug('Cancelling sub {}'.format(dependent_job))
                        self.cancel_job(dependent_job)
            else:
                fastr.log.debug('No children....')

    def hold_job(self, job):
        if self.SUPPORTS_HOLD_RELEASE:
            if not isinstance(job, Job):
                if job not in self.job_dict:
                    fastr.log.warning('Job {} is no longer under processing, cannot release!'.format(job))
                    return

                try:
                    job = self.job_dict[job]
                except KeyError:
                    fastr.log.warning('Job {} is no longer under processing, cannot release!'.format(job))
                    return

            job.status = JobState.hold
            self._hold_job(job)
        else:
            raise exceptions.FastrNotImplementedError('Cannot handle hold/release by default yet!')

    def release_job(self, job):
        """
        Release a job that has been put on hold

        :param job: job to release
        """
        if not isinstance(job, Job):
            try:
                job = self.job_dict[job]
            except KeyError:
                fastr.log.warning('Job {} is no longer under processing, cannot release!'.format(job))
                return

        job.status = JobState.queued
        if self.SUPPORTS_HOLD_RELEASE:
            # Job is already queued, but held so it need to be released
            self._release_job(job)
        else:
            # Job was never queued in the first place
            self.queue_job(job)

    def job_finished(self, job, errors=None, blocking=False):
        """
        The default callback that is called when a Job finishes. This will
        create a new thread that handles the actual callback.

        :param Job job: the job that finished
        :param errors: optional list of errors encountered
        :param bool blocking: if blocking, do not create threads
        :return:
        """
        if isinstance(job, SourceJob):
            if len(self.source_job_queue) > 0:
                fastr.log.debug('Taking source job from the queue')
                new_source_job = self.source_job_queue.popleft()
                if self.SUPPORTS_HOLD_RELEASE:
                    self.release_job(new_source_job)
                else:
                    self._dispatch_job(new_source_job)
            else:
                fastr.log.debug('Releasing source job slot {} -> {}'.format(self.number_of_source_jobs_queued,
                                                                            self.number_of_source_jobs_queued - 1))
                self.number_of_source_jobs_queued -= 1

        if not blocking:
            # The callback has to finish immediately, so create thead to handle callback and return
            callback_thread = threading.Thread(target=self._job_finished_body,
                                               name='fastr_jobfinished_callback',
                                               args=(job, errors))
            callback_thread.start()
        else:
            self._job_finished_body(job, errors)

    def _job_finished_body(self, job, errors):
        """
        The actual callback that is executed in a separate thread. This
        method handles the collection of the result, the release of depending
        jobs and calling the user defined callback.

        :param Job job: the job that finished
        :param errors: optional list of errors encountered
        """
        fastr.log.debug('ExecutorInterface._job_finished_callback called for {}'.format(job))
        self.job_status[job.id] = JobState.processing_callback

        # The Job finished should always log the errors rather than
        # crashing the whole execution system
        # pylint: disable=bare-except
        try:
            try:
                job = iohelpers.load_gpickle(job.logfile)
            except EOFError:
                job.errors.append(
                    exceptions.FastrResultFileNotFound(
                        ('Could not read job result file {}, assuming '
                         'the job crashed during output write.').format(job.logfile)).excerpt())
                job.status = JobState.failed
            except IOError:
                job.errors.append(
                    exceptions.FastrResultFileNotFound(
                        ('Could not find/read job result file {}, assuming '
                         'the job crashed before it created output.').format(job.logfile)).excerpt())
                job.status = JobState.failed

            if self._status_callback is not None:
                self._status_callback(job)
                job.status_callback = self._status_callback

        except:
            exc_type, _, trace = sys.exc_info()
            exc_info = traceback.format_exc()
            trace = traceback.extract_tb(trace, 1)[0]
            fastr.log.error('Encountered exception ({}) during execution:\n{}'.format(exc_type.__name__, exc_info))
            job.errors.append((exc_type.__name__, exc_info, trace[0], trace[1]))
            job.status = JobState.execution_failed

        # The execution plugin found that there were errors, append them
        if errors is not None and len(errors) > 0:
            job.errors.extend(errors)

            # Resave the job with the updated error list
            fastr.log.info('Found errors from execution plugin, updating pickle {}'.format(job.logfile))
            iohelpers.save_gpickle(job.logfile, job)

        # Strip provenance information from the job
        del job.provenance

        result = job
        fastr.log.debug('Finished {} with status {}'.format(job.id, job.status))
        job_id = result.id

        # Make sure the status is either finished or failed
        if result.status == JobState.execution_done:
            result.status = JobState.finished
        else:
            result.status = JobState.failed

        # Set the job status so the hold jobs will be release properly
        self.job_status[job_id] = result.status
        if job_id in self.job_dict:
            self.job_dict[job_id].status = result.status

        # Do the callbacks before releasing the other jobs
        # Extra subclass callback
        fastr.log.debug('Subclass callback')
        try:
            self._job_finished(result)
        except:
            exc_type, _, _ = sys.exc_info()
            exc_info = traceback.format_exc()
            fastr.log.error('Encountered exception ({}) during callback {}._job_finished:\n{}'.format(exc_type.__name__, type(self).__name__, exc_info))

        # Extra callback from object
        fastr.log.debug('Calling callback for {}'.format(job_id))
        if self._finished_callback is not None:
            try:
                self._finished_callback(result)
            except:
                if isinstance(self._finished_callback, functools.partial):
                    args = self._finished_callback.args + tuple('{}={}'.format(k, v) for k, v in self._finished_callback.keywords.items())
                    callback_name = '{f.__module__}.{f.func_name}({a})'.format(f=self._finished_callback.func, a=','.join(args))
                elif isinstance(self._finished_callback, types.FunctionType):
                    callback_name = '{f.__module__}.{f.func_name}'.format(f=self._finished_callback)
                elif isinstance(self._finished_callback, types.MethodType):
                    callback_name = '{m.__module__}.{m.im_class.__name__}.{m.im_func.func_name}'.format(m=self._finished_callback)
                else:
                    callback_name = repr(self._finished_callback)
                exc_type, _, _ = sys.exc_info()
                exc_info = traceback.format_exc()
                fastr.log.error('Encountered exception ({}) during callback {}:\n{}'.format(exc_type.__name__, callback_name, exc_info))
        else:
            fastr.log.debug('No callback specified')

        # The ProcessPoolExecutor has to track job dependencies itself, so
        # therefor we have to check for jobs depending on the finished job
        if job_id in self.held_queue:
            fastr.log.debug('Signaling depending jobs {}'.format(self.held_queue[job_id]))
            ready_jobs = []
            for held_job_id in self.held_queue[job_id]:
                action = self.check_job_requirements(held_job_id)
                if action == JobAction.queue:
                    # Re-assign managed dict member
                    if held_job_id in self.job_dict:
                        held_job = self.job_dict[held_job_id]
                        self.job_dict[held_job_id] = held_job
                    fastr.log.debug('Job {} is now ready to be submitted'.format(held_job_id))

                    # If ready, flag job for removal from held queue and send
                    # to pool queue to be executed
                    ready_jobs.append(held_job_id)
                    if not self.SUPPORTS_DEPENDENCY:
                        self.release_job(held_job_id)
                elif action == JobAction.cancel:
                    fastr.log.debug('Job {} will be cancelled'.format(held_job_id))
                    ready_jobs.append(held_job_id)
                    self.cancel_job(held_job_id)
                else:
                    fastr.log.debug('Job {} still has unmet dependencies'.format(held_job_id))

            # Remove jobs that no longer need to be held from held_queue
            for ready_job_id in ready_jobs:
                job = self.get_job(ready_job_id)

                for hold_id in job.hold_jobs:
                    if hold_id in self.held_queue:
                        with self.held_queue_lock:
                            # remove from held_queue. because of the managed dict,
                            # we need to replace the value (not update)
                            required_job = self.held_queue[hold_id]
                            if ready_job_id in required_job:
                                required_job.remove(ready_job_id)
                            else:
                                fastr.log.warning('Could not remove {} from dependencies {} of {}'.format(ready_job_id, required_job, hold_id))

            with self.held_queue_lock:
                del self.held_queue[job_id]

        # Move the job to archive (to keep the number of working jobs limited
        # in the future the archive can be moved to a db/disk if needed
        fastr.log.debug('Archiving job {} with status {}'.format(job_id, result.status))
        try:
            del self.job_status[job_id]
        except KeyError:
            pass

        self.job_archive[job_id] = result

        try:
            del self.job_dict[job_id]
        except KeyError:
            pass
        fastr.log.debug('Done archiving')

    def get_job(self, job_id):
        try:
            return self.job_dict[job_id]
        except KeyError:
            try:
                return self.job_archive[job_id]
            except:
                raise exceptions.FastrKeyError('Could not find job {}'.format(job_id))

    def get_status(self, job):
        if not isinstance(job, Job):
            job = self.get_job(job)

        return job.status

    @abstractmethod
    def _queue_job(self, job):
        """
        Method that a subclass implements to actually queue a Job for execution

        :param job: job to queue

        .. note:: If ``SUPPORT_HOLD_RELEASE=True`` the plugin should check
                  ``job.status`` to see if it is ``JobStatus.hold``. If
                  this is the case it should submit the job in a held
                  state to be released at a later time. If
                  ``SUPPORT_HOLD_RELEASE=False`` the status be ignored.
        """

    def _cancel_job(self, job):
        """
        Method that a subclass implements to actually cancel a Job

        :param job: job to cancel
        """
        if self.SUPPORTS_CANCEL:
            raise exceptions.FastrPluginCapabilityNotImplemented('The plugin should have reimplemented _cancel_job!')
        else:
            fastr.log.warning('Cannot remove job {} from a {} queue!'.format(job.id,
                                                                             type(self).__name__))

    def _hold_job(self, job):
        """
        Method that a subclass implements to hold a queued/running Job, only
        required if ``SUPPORT_HOLD_RELEASE=True``

        :param job: Job to hold
        """
        if self.SUPPORTS_HOLD_RELEASE:
            raise exceptions.FastrPluginCapabilityNotImplemented(
                'Cannot hold job {}: the plugin should have reimplemented _hold_job!'.format(
                    job.id
                )
            )

    def _release_job(self, job):
        """
        Method that a subclass implements to actually release a job, only
        required if ``SUPPORT_HOLD_RELEASE=True``

        :param job: job to release from hold
        """
        if self.SUPPORTS_HOLD_RELEASE:
            raise exceptions.FastrPluginCapabilityNotImplemented(
                'Cannot release job {}: the plugin should have reimplemented _release_job!'.format(
                    job.id
                )
            )

    def _job_finished(self, job):
        """
        Method that a subclass can implement to add to the default callback.
        It will be called by ``_job_finished_body`` right before the user
        defined callback will be called.

        :param job: Job that resulted from the execution
        """

    def show_jobs(self, req_status=None):
        """
        List the queued jobs, possible filtered by status

        :param req_status: requested status to filter on
        :return: list of jobs
        """
        if isinstance(req_status, basestring):
            req_status = JobState[req_status]

        if not isinstance(req_status, JobState):
            return []

        results = []
        for key, status in self.job_status.items():
            if req_status is None or status == req_status:
                results.append(self.get_job(key))

        return results

    def check_job_status(self, job_id):
        """
        Get the status of a specified job

        :param job_id: the target job
        :return: the status of the job (or None if job not found)
        """
        try:
            return self.get_status(job_id)
        except exceptions.FastrKeyError:
            return None

    def check_job_requirements(self, job_id):
        """
        Check if the requirements for a job are fulfilled.

        :param job_id: job to check
        :return: directive what should happen with the job
        :rtype: JobAction
        """
        job = self.get_job(job_id)
        if job.hold_jobs is None or len(job.hold_jobs) == 0:
            return JobAction.queue

        status_list = [self.check_job_status(job_id) for job_id in job.hold_jobs]

        if all(x.done for x in status_list):
            if all(x == JobState.finished for x in status_list):
                return JobAction.queue
            else:
                return JobAction.cancel
        else:
            return JobAction.hold


class ExecutionPluginManager(PluginSubManager):
    """
    Container holding all the ExecutionPlugins known to the Fastr system
    """

    def __init__(self):
        """
        Initialize a ExecutionPluginManager and load plugins.

        :param path: path to search for plugins
        :param recursive: flag for searching recursively
        :return: newly created ExecutionPluginManager
        """
        super(ExecutionPluginManager, self).__init__(parent=fastr.plugin_manager,
                                                     plugin_class=ExecutionPlugin)




#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys
import os
import imp
import StringIO

FASTR_LOG_TYPE = 'none'  # Do not get info about fastr

import fastr
from fastr.core.tool import Tool
from fastr.core.version import Version


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('infile', metavar='SCRIPT.py', help='Python script to inspect')
    parser.add_argument('outfile', metavar='TOOL.xml', help='created Tool stub')
    return parser


def cardinality_from_nargs(value):
    val_mapping = {'+': '1-*',
                   '?': '0-1',
                   '*': '0-*',
                   None: 1}

    if isinstance(value, int):
        return value
    else:
        return val_mapping[value]


def datatype_from_type(type_, metavar):
    type_mapping = {int: 'Int',
                    float: 'Float',
                    bool: 'Boolean',
                    str: 'String',
                    None: 'String'}
    if type_ in (None, str) and isinstance(metavar, str):
        return fastr.typelist.guess_type(metavar.lower(), exists=False).id
    else:
        if type_ in (None, str) and metavar is not None:
            type_ = type(metavar)
        return type_mapping.get(type_, 'AnyType')


def main():
    """
    Create a stub for a Tool based on a python script using argparse
    """
    # No arguments were parsed yet, parse them now
    parser = get_parser()
    args = parser.parse_args()

    # Extract arguments
    filepath = args.infile
    output_path = args.outfile

    # Extract argparse from desired script
    parser, stdout = extract_argparser(filepath)
    parser.prog = os.path.basename(filepath)

    description = parser.description
    if description is None:
        description = 'auto generated tool from {}'.format(filepath)

    tool = Tool()
    tool.id = os.path.splitext(os.path.basename(filepath))[0]
    tool.name = tool.id
    tool.version = Version('1.0')
    tool.filename = output_path
    tool.node_class = 'Node'
    tool.command = {'authors': [],
                    'description': description,
                    'targets': [{'os': '*',
                                 'arch': '*',
                                 'bin': os.path.basename(filepath),
                                 'interpreter': 'python',
                                 'location': fastr.vfs.path_to_url(os.path.dirname(filepath)),
                                 'module': None}],
                    'url': 'http://www.bigr.nl/fastr',
                    'version': Version('1.0'),
                    'license': 'UNKNOWN'}
    tool.help = parser.format_help()

    tool.authors = [{'name': 'extract_argparse.py',
                     'email': 'jsmith@example.org',
                     'url': 'http://example.org'}]
    tool.description = description
    tool.tests = []

    inputs = []

    for option in parser._optionals._group_actions:
        prefix = max(option.option_strings, key=len)
        id_ = prefix.strip('-')
        param = {'id': id_,
                 'name': id_,
                 'prefix': prefix,
                 'description': option.help,
                 'cardinality': cardinality_from_nargs(option.nargs),
                 'required': option.required,
                 'repeat_prefix': False}

        if option.const is not None:
            param['datatype'] = datatype_from_type(option.type, option.const)
        elif option.choices is not None:
            param['enum'] = [str(x) for x in option.choices]
        else:
            param['datatype'] = datatype_from_type(option.type, option.metavar)

        inputs.append(param)

    tool.interface = fastr.plugins.FastrInterface(id_='{}_{}_interface'.format(tool.id, tool.command_version),
                                                  document={
                                                      'inputs': inputs,
                                                      'outputs': [],
                                                  })

    if os.path.isdir(output_path):
        output_path = os.path.join(output_path,
                                   tool.id + '.xml')
    with open(output_path, 'w') as output_fh:
        output_fh.write(tool.dumps('xml'))
    print('Written Tool stub to {}...'.format(output_path))


def extract_argparser(filepath):
    name = os.path.splitext(os.path.basename(filepath))[0]
    mod = imp.load_source(name, filepath)
    if not hasattr(mod, 'main'):
        print('Module does not have a main() function')
        return

    return find_argparser(mod.main, filepath)


def find_argparser(entry, basename=sys.argv[0]):
    # Catch stdout to string (no auto-print)
    old_stdout = sys.stdout
    temp_buffer = StringIO.StringIO()
    sys.stdout = temp_buffer

    try:
        # Call entry with --help to trigger help and SystemExit by argparse
        sys.argv = [basename, '--help']
        entry()
    except SystemExit:
        # Follow traceback frames until we find the first ArgumentParser object
        tb = sys.exc_info()[2]
        tb_stack = [tb]
        while tb.tb_next is not None:
            tb = tb.tb_next
            tb_stack.append(tb)
        for tb in tb_stack:
            for name, obj in tb.tb_frame.f_locals.items():
                if isinstance(obj, argparse.ArgumentParser):
                    return obj, temp_buffer.getvalue()
        raise
    finally:
        # Cleanup StringIO and reset stdout
        sys.stdout = old_stdout
        temp_buffer.close()


if __name__ == '__main__':
    main()




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import fastr
from fastr import exceptions
from fastr.data import url
from fastr.plugins import FastrInterface


class PathCollector(FastrInterface.collector_plugin_type):
    """
    The PathCollector plugin for the FastrInterface. This plugin uses the
    location fields to find data on the filesystem. To use this plugin the
    method of the output has to be set to ``path``

    The general working is as follows:

    1. The location field is taken from the output
    2. The substitutions are performed on the location field (see below)
    3. The updated location field will be used as a :ref:`regular expression <python:re-syntax>` filter
    4. The filesystem is scanned for all matching files/directory

    The special substitutions performed on the location use the
    Format Specification Mini-Language :ref:`python:formatspec`.
    The predefined fields that can be used are:

    * ``inputs``, an objet with the input values (use like ``{inputs.image[0]}``)
      The input contains the following attributes that you can access:

      * ``.directory`` for the directory name (use like ``input.image[0].directory``)
        The directory is the same as the result of ``os.path.dirname``
      * ``.filename`` is the result of ``os.path.basename`` on the path
      * ``.basename`` for the basename name (use like ``input.image[0].basename``)
        The basename is the same as the result of ``os.path.basename`` and the
        extension stripped. The extension is considered to be everything after
        the first dot in the filename.
      * ``.extension`` for the extension name (use like ``input.image[0].extension``)

    * ``output``, an object with the output values (use like ``{outputs.result[0]}``)
      It contains the same attributes as the input

      * ``special.cardinality``, the index of the current cardinality
      * ``special.extension``, is the extension for the output DataType

    Example use::

      <output ... method="path" location="{output.directory[0]}/TransformParameters.{special.cardinality}.{special.extension}"/>

    Given the output directory ``./nodeid/sampleid/result``, the second sample in the output and
    filetype with a ``txt`` extension, this would be translated into::

      <output ... method="path" location="./nodeid/sampleid/result/TransformParameters.1.txt>

    """
    def __init__(self):
        super(PathCollector, self).__init__()
        self.id = 'path'

    def _collect_results(self, interface, output, result):
        output_data = []
        result.result_data[output.id] = output_data
        location = output.location
        use_cardinality = re.search(r'\{special.cardinality(:.*?)??\}', location) is not None

        if use_cardinality:
            nr = 0

            while True:
                specials, inputs, outputs = interface.get_specials(result.payload, output, nr)

                fastr.log.debug('input: {}, output: {}, specials: {}'.format(inputs, outputs, specials))
                fastr.log.debug('location: {}'.format(location))
                fastr.log.debug('parsed location: {}'.format(location.format(input=inputs,
                                                                             output=outputs,
                                                                             special=specials,
                                                                             input_parts=inputs,
                                                                             output_parts=outputs)))

                value = location.format(input=inputs,
                                        output=outputs,
                                        special=specials,
                                        input_parts=inputs,
                                        output_parts=outputs)
                fastr.log.debug('Searching regexp value {}'.format(value))

                if url.isurl(value):
                    value = fastr.vfs.url_to_path(value)

                path_list = self._regexp_path(value)

                if len(path_list) < 1:
                    break

                if len(path_list) > 1:
                    message = 'Found multiple matches for automatic output using {}'.format(value)
                    fastr.log.error(message)
                    raise exceptions.FastrCollectorError(message)

                value = path_list[0]

                fastr.log.debug('Got automatic result: {}'.format(value))
                output_data.append(value)

                nr += 1

        else:
            fastr.log.debug('No cardinality branch!')
            specials, inputs, outputs = interface.get_specials(result.payload, output, '')
            value = location.format(input=inputs,
                                    output=outputs,
                                    special=specials,
                                    input_parts=inputs,
                                    output_parts=outputs)
            if url.isurl(value):
                value = fastr.vfs.url_to_path(value)

            path_list = self._regexp_path(value)

            output_data.extend(path_list)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains both the Updateable class and UpdateableMeta meta-class
for objects which support updates within the fastr system
"""

from abc import ABCMeta, abstractmethod
import os
import types
import threading
import fastr

import fastr.exceptions as exceptions


class UpdateableMeta(ABCMeta):
    """
    A metaclass for objects which are updateable and need some methods/properties
    to trigger an update.
    """

    @classmethod
    def calcmro(mcs, bases):
        """Calculate the Method Resolution Order of bases using the C3 algorithm.

        Suppose you intended creating a class K with the given base classes. This
        function returns the MRO which K would have, *excluding* K itself (since
        it doesn't yet exist), as if you had actually created the class.

        Another way of looking at this, if you pass a single class K, this will
        return the linearization of K (the MRO of K, *including* itself).

        :param bases: the list of bases for which create the MRO
        :return: the list representing the entire MRO, except the (non-existing) class itself

        Note: Taken from http://code.activestate.com/recipes/577748-calculate-the-mro-of-a-class/
              Created by Steven D'Aprano and licensed under the MIT license
        """
        seqs = [list(C.__mro__) for C in bases] + [list(bases)]
        res = []
        while True:
            non_empty = [item for item in seqs if item]
            if not non_empty:
                # Nothing left to process, we're done.
                return tuple(res)
            for seq in non_empty:  # Find merge candidates among seq heads.
                candidate = seq[0]
                not_head = [s for s in non_empty if candidate in s[1:]]
                if not_head:
                    # Reject the candidate.
                    candidate = None
                else:
                    break
            if not candidate:
                raise TypeError("inconsistent hierarchy, no C3 MRO is possible")
            res.append(candidate)
            for seq in non_empty:
                # Remove candidate.
                if seq[0] == candidate:
                    del seq[0]

    @classmethod
    def find_member(mcs, name, parents, dct):
        """
        Find  a member of the class in the same way as Python would if it had a given dict and set of bases

        :param mcs: metaclass at work
        :param name: name of the class to be created
        :param parents: list of the bases for the new class
        :param dct: the dict of the class being created
        :return: the firstly resolved member or None if nothing found
        """
        if name in dct:
            return dct[name]

        for cls in mcs.calcmro(parents):
            if hasattr(cls, name):
                return getattr(cls, name)

        return None

    def __new__(mcs, name, parents, dct):
        if '__updatetriggers__' in dct:
            triggers = dct['__updatetriggers__']

            for trigger in triggers:
                fnc = mcs.find_member(trigger, parents, dct)

                if fnc is not None:
                    if isinstance(fnc, (types.FunctionType, types.MethodType, types.BuiltinMethodType)):
                        fastr.log.debug('Adding update trigger {} to {}'.format(trigger, name))
                        dct[trigger] = mcs.updatetrigger(fnc)
                    else:
                        fastr.log.debug('Skipping trigger {} for {} (wrong type: {})'.format(trigger, name, type(fnc).__name__))
                else:
                    fastr.log.debug('Skipping trigger {} for {} (not in dct)'.format(trigger, name))
                    raise ValueError()

        return super(UpdateableMeta, mcs).__new__(mcs, name, parents, dct)

    @staticmethod
    def updatetrigger(fnc):
        """
        Function decorator to make a function trigger an update after being
        called. This is a way to easily have function trigger an update
        after setting a value without writing tons of wrapper functions.
        The function keeps the original docstring and appends a note to it.
        """
        def wrapper(self, *args, **kwargs):
            """
            Decorator wrapper around a function. This docstring will be
            changed for each wrapped function
            """
            # First call the wrapped function
            fnc(self, *args, **kwargs)
            # Call the update function
            # This function name is chosen so that it has low chance of conflicts
            if self.__updating__:
                self.__updatefunc__()

        # Add a note to the docstring indicating what is going on
        original = '{}.{}'.format(fnc.__module__, fnc.__name__)
        docstring = fnc.__doc__.lstrip('\n')
        indent = docstring[:len(docstring) - len(docstring.lstrip())]
        extra_doc = ('\n.. note::\n'
                     '    This is a wrapped version of ``{orig}``\n'
                     '    which triggers an update of  the object after\n'
                     '    being called').format(orig=original)
        extra_doc = '\n{}'.format(indent).join(extra_doc.splitlines())
        wrapper.__doc__ = docstring.rstrip() + "\n{indent}\n{indent}{extra_doc}\n{indent}".format(indent=indent, extra_doc=extra_doc)
        return wrapper


class Updateable(object):

    """
    Super class for all classes that can be updated and have a status.
    These objects can be valid/invalid and ready/not-ready depending on their
    state. These states are set by the function update. This allows for
    interactively checking the network.
    """

    __metaclass__ = UpdateableMeta

    #: Which methods need to be wrapped to trigger an update. Override this
    #: value to have the functions automatically wrapped. E.g.
    #: ``__update_triggers__ = ['append', 'insert', '__setitem__']`` to have
    #: these functions wrapped.
    __updatetriggers__ = []

    #: Flag to indicate that this object is allowed to update
    __updating__ = True

    #: Lock to avoid multiple updates happening at the same time
    __updateinprogress__ = threading.Lock()

    def __init__(self):
        """
        Constructor, creates the status field

        :return: newly created object
        """
        self._status = {'key': None, 'ready': None, 'valid': None, 'messages': []}

    def __getstate__(self):
        """
        Retrieve the state of the object, make sure the status is not part of
        the description as it will not be valid after re-creating the object.

        :return: the state of the object
        :rtype dict:
        """
        return {}

    def __setstate__(self, state):
        """
        Set the state of the object by the given state. This adds a clean
        status field, making sure it is not unintended, outdated information
        from before serialization.

        :param dict state: The state to populate the object with
        """
        self._status = {'key': None, 'ready': None, 'valid': None, 'messages': []}

    @property
    def messages(self):
        """
        The messages of the last update
        """
        return self._status['messages']

    @property
    def ready(self):
        """
        Flag indicating that the object is ready
        """
        return self._status['ready']

    @property
    def valid(self):
        """
        Flag indicating that the object is valid
        """
        return self._status['valid']

    def update(self, key=None, forward=True, backward=False):
        """
        Default function for updating, it can be called without key to have a
        new update started with a new key.

        :param int key: a key for this update, should be different than the
                        last update key
        :param bool forward: flag indicating to update forward in the network
        :param bool backward: flag indicating to update backward in the network
        """
        # If updating is disable, we don't do anything
        if not self.__updating__:
            return

        # If this is the first update function called in this round, create a key
        lock = False
        if key is None:
            key = hash(os.urandom(128))

            # Block update until previous updates are finished
            fastr.log.debug('Getting update lock')
            lock = Updateable.__updateinprogress__.acquire()
            fastr.log.debug('Got update lock')

        # Only update if not update in this update round
        if key == self._status['key']:
            # Release updatable lock, this should never be needed, but if not done when needed causes a deadlock
            if lock:
                fastr.log.warning('Releasing lock in unexpected place! Might be a bug!')
                Updateable.__updateinprogress__.release()
                fastr.log.debug('Released update lock')
            return

        # Store key to mark last update round
        self._status['key'] = key

        # Run the update func
        self._update(key, forward, backward)

        # Release updatable lock
        if lock:
            Updateable.__updateinprogress__.release()
            fastr.log.debug('Released update lock')

        return key

    # Set the update function for the trigger functions to call
    __updatefunc__ = update

    @abstractmethod
    def _update(self, key, forward=True, backward=False):
        """
        The actual update function to be used. This is an abstract function
        that Updateable objects must implement.

        :param int key: a key for this update, should be different than the
                        last update key
        :param bool forward: flag indicating to update forward in the network
        :param bool backward: flag indicating to update backward in the network
        """
        raise exceptions.FastrNotImplementedError('Purposefully not implemented')



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imp
import os
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

with open('requirements.txt', 'r') as fh:
    _requires = fh.read().splitlines()

with open('test_requirements.txt', 'r') as fh:
    _tests_require = fh.read().splitlines()

with open('README', 'r') as fh:
    _description = fh.read()


def scan_dir(path, prefix=None):
    if prefix is None:
        prefix = path

    # Scan resources package for files to include
    file_list = []
    for root, dirs, files in os.walk(path):
        # Strip this part as setup wants relative directories
        root = root.replace(prefix, '')
        root = root.lstrip('/\\')

        for filename in files:
            if filename[0:8] == '__init__':
                continue
            file_list.append(os.path.join(root, filename))

    return file_list

# Create version file
version = imp.load_source('version', os.path.join(os.path.dirname(__file__), 'fastr', 'version.py'))
__version__ = version.version
version.save_version(__version__, version.hg_head, version.hg_branch)

# Determine the extra resources and scripts to pack
resources_list = scan_dir('./fastr/resources')
examples_list = scan_dir('./fastr/examples/data', './fastr/examples')
web_list = scan_dir('./fastr/web/static', './fastr/web')
web_list += scan_dir('./fastr/web/templates', './fastr/web')

print('[setup.py] called with: {}'.format(' '.join(sys.argv)))
if hasattr(sys, 'real_prefix'):
    print('[setup.py] Installing in virtual env {} (real prefix: {})'.format(sys.prefix, sys.real_prefix))
else:
    print('[setup.py] Not inside a virtual env!')


class NoseTestCommand(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # Run nose ensuring that argv simulates running nosetests directly
        import nose
        nose.run_exit(argv=['nosetests'])

# Set the entry point
entry_points = {
    "console_scripts": [
        "fastr = fastr.utils.cmd.__init__:main",
    ]
}

setup(
    name='fastr',
    version=__version__,
    author='H.C. Achterberg, M. Koek',
    author_email='hakim.achterberg@gmail.com',
    packages=['fastr',
              'fastr.core',
              'fastr.core.test',
              'fastr.data',
              'fastr.examples',
              'fastr.examples.test',
              'fastr.execution',
              'fastr.resources',
              'fastr.utils',
              'fastr.utils.cmd',
              'fastr.web'],
    package_data={'fastr.resources': resources_list,
                  'fastr.examples': examples_list,
                  'fastr.web': web_list,
                  'fastr': ['versioninfo']
                  },
    scripts=[],
    url='https://bitbucket.org/bigr_erasmusmc/fastr',
    license='Apache License 2.0',
    description='Workflow creation and batch execution environment.',
    long_description=_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Education',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Logging',
        'Topic :: Utilities',
    ],
    install_requires=_requires,
    tests_require=_tests_require,
    test_suite='nose.collector',
    cmdclass={'test': NoseTestCommand},
    entry_points=entry_points,
)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module to generate reStructuredText tables from json schema files
"""

from jsonschema import RefResolver


class SchemaPrinter(object):
    """
    Object that create a table in reStructuedText from a json schema
    """
    def __init__(self, schema, skipfirst=False):
        """
        Create the printer object

        :param dict schema: the json schema to print
        :param bool skipfirst: flag to indicate that the first line should not be printed
        """
        self.schema = schema
        self.location = []
        self.depth = 0
        self.maxdepth = 0
        self.length = {}
        self.description_length = 12
        self.lines = []
        self.resolver = RefResolver('', self.schema)
        self.skipfirst = skipfirst
        self.parse()

    def __str__(self):
        """
        String representation of json schema (that is the printed table)
        """
        return '\n'.join(self.printlines())

    def descend(self, properties):
        """
        Descend into a subschema

        :param dict properties: the properties in the subschema
        """
        for key, subschema in sorted(properties.items()):
            key = '``{}``'.format(key)
            self.location.append(key)
            self.depth += 1
            self.maxdepth = max(self.maxdepth, self.depth)
            if self.depth not in self.length:
                self.length[self.depth] = len(key) + 2
            else:
                self.length[self.depth] = max(self.length[self.depth], len(key) + 2)

            self.parse(subschema)

            self.location.pop()
            self.depth -= 1

    def parse(self, schema=None):
        """
        Parse a schema

        :param dict schema: the schema to parse
        """
        if schema is None:
            schema = self.schema

        # If it is a reference, follow and write with that instead
        if '$ref' in schema:
            with self.resolver.resolving(schema['$ref']) as newschema:
                return self.parse(newschema)

        if 'type' not in schema:
            return

        datatype = schema['type']

        if 'description' in schema:
            self.description_length = max(self.description_length, len(schema['description']))
            description = schema['description']
        else:
            description = ''

        if datatype == 'array':
            self.location[-1] = '``{}[]``'.format(self.location[-1][2:-2])

        self.lines.append((self.location[:], datatype, description))

        if datatype == 'object':
            self.descend(schema['properties'])
        elif datatype == 'array':
            if isinstance(schema['items'], dict):
                items = schema['items']
                if '$ref' in items:
                    with self.resolver.resolving(items['$ref']) as newschema:
                        if 'properties' in  newschema:
                            return self.descend(newschema['properties'])

                elif 'properties' in items:
                    self.descend(items['properties'])

    def printlines(self):
        """
        Given a parsed schema (parsing happens when the object is constructed), print all the lines

        :return: the printed table
        :rtype: str
        """
        output_lines = []
        sepertor = ''.join('+{{keys[{n}]:{{fill[{n}]}}<{length}}}'.format(length=l, n=nr-1) for nr, l in self.length.items())
        sepertor = '{start}+{{desc:{{dfill}}<{length}}}+'.format(start=sepertor, length=self.description_length)
        contentline = ''.join('{{sep[{n}]}}{{keys[{n}]: <{length}}}'.format(length=l, n=nr-1) for nr, l in self.length.items())
        contentline = '{start}|{{desc: <{length}}}|'.format(start=contentline, length=self.description_length)
        lastkey = []
        dummykey = [''] * self.maxdepth
        first = True
        if self.skipfirst:
            lines = self.lines[1:]
        else:
            lines = self.lines

        for line in lines:
            key = line[0]
            limit = min(len(lastkey), len(key))
            same = [key[k] == lastkey[k] if k < limit else False for k in range(self.maxdepth)]
            fill = [' ' if x else '-' for x in same]
            keys = [key[k] if k < len(key) and not same[k] else '' for k in range(self.maxdepth)]
            sep = ['|' if k <= len(key) else ' ' for k in range(self.maxdepth)]
            output_lines.append(sepertor.format(keys=dummykey, fill=fill, desc='', dfill='-'))
            if first:
                headerkey = dummykey[:]
                headerkey[0] = 'Attributes'
                headersep = '|' + ' ' * (self.maxdepth - 1)
                output_lines.append(contentline.format(keys=headerkey, desc='Description', sep=headersep))
                output_lines.append(sepertor.format(keys=dummykey, fill='===', desc='', dfill='='))
            output_lines.append(contentline.format(keys=keys, desc=line[2], sep=sep))
            first = False
            lastkey = key
        fill = '-' * self.maxdepth
        output_lines.append(sepertor.format(keys=dummykey, fill=fill, desc='', dfill='-'))
        return output_lines




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the VirtualFileSystemRegularExpression plugin for fastr
"""

import os
import re
import urlparse
import fastr
import fastr.exceptions as exceptions
from fastr.core.ioplugin import IOPlugin


class VirtualFileSystemRegularExpression(IOPlugin):
    """
    The VirtualFileSystemValueList an expand-only type of IOPlugin. No URLs
    can actually be fetched, but it can expand a single URL into a larger
    amount of URLs.

    A ``vfsregex://`` URL is a vfs URL that can contain regular expressions
    on every level of the path. The regular expressions follow the
    :mod:`re module <re>` definitions.

    An example of a valid URLs would be::

        vfsregex://tmp/network_dir/.*/.*/__fastr_result__.pickle.gz
        vfsregex://tmp/network_dir/nodeX/(?P<id>.*)/__fastr_result__.pickle.gz

    The first URL would result in all the ``__fastr_result__.pickle.gz`` in
    the working directory of a Network. The second URL would only result in
    the file for a specific node (nodeX), but by adding the named group
    ``id`` using ``(?P<id>.*)`` the sample id of the data is automatically
    set to that group (see :ref:`python:re-syntax` under the special
    characters for more info on named groups in regular expression).

    Concretely if we would have a directory ``vfs://mount/somedir`` containing::

        image_1/Image.nii
        image_2/image.nii
        image_3/anotherimage.nii
        image_5/inconsistentnamingftw.nii

    we could match these files using ``vfsregex://mount/somedir/(?P<id>image_\\d+)/.*\\.nii``
    which would result in the following source data after expanding the URL::

        {'image_1': 'vfs://mount/somedir/image_1/Image.nii',
         'image_2': 'vfs://mount/somedir/image_2/image.nii',
         'image_3': 'vfs://mount/somedir/image_3/anotherimage.nii',
         'image_5': 'vfs://mount/somedir/image_5/inconsistentnamingftw.nii'}

    Showing the power of this regular expression filtering. Also it shows how
    the ID group from the URL can be used to have sensible sample ids.

    .. warning:: due to the nature of regexp on multiple levels, this method
                 can be slow when having many matches on the lower level of
                 the path (because the tree of potential matches grows) or
                 when directories that are parts of the path are very large.
    """
    scheme = 'vfsregex'

    def __init__(self):
        # initialize the instance and register the scheme
        super(VirtualFileSystemRegularExpression, self).__init__()

    def expand_url(self, url):
        if fastr.data.url.get_url_scheme(url) != 'vfsregex':
            raise exceptions.FastrValueError('URL not of vfsregex type!')

        # Retrieve base path
        parsed = urlparse.urlparse(url)
        baseurl = urlparse.urlunparse(urlparse.ParseResult(scheme='vfs', netloc=parsed.netloc, path='', params='', query='', fragment=''))
        basepath = fastr.ioplugins.url_to_path(baseurl)

        # Add the "query" back to the path, as we do not want a regex ? to trigger an end in the path
        if parsed.query != '':
            path = parsed.path + '?' + parsed.query
        else:
            path = parsed.path

        subpaths = fastr.data.url.full_split(path)
        if subpaths[0] != '/':
            return ValueError('vfs url should always contain a valid path')

        pathlist = [(None, basepath)]
        fastr.log.debug('Basepath: {}, Subpaths {}'.format(basepath, subpaths))
        for subpath in subpaths[1:]:
            subpath = '^{}$'.format(subpath)

            # Test the regexp and give a more understandable error if it fails
            try:
                re.compile(subpath)
            except Exception as detail:
                raise exceptions.FastrValueError('Error parsing regexp "{}": {}'.format(subpath, detail))

            # Scan new level for matches
            newpathlist = []
            for id_, curpath in pathlist:
                contents = os.listdir(curpath)
                for option in contents:
                    match = re.match(subpath, option)
                    if match is not None:
                        try:
                            id_ = match.group('id')
                        except IndexError:
                            pass

                        newpathlist.append((id_, os.path.join(curpath, option)))

            pathlist = newpathlist

        pathlist = tuple((id_, re.sub(basepath, baseurl + '/', x)) for id_, x in sorted(pathlist))
        return pathlist



import os
import gzip

import fastr
from fastr.core.serializable import Serializable
from fastr.core.tool import Tool


def verify_resource_loading(filename):
    name, ext = os.path.splitext(filename)

    # Check if file is gzipped
    if ext == '.gz':
        compressed = True
        name, ext = os.path.splitext(filename)
    else:
        compressed = False

    # Read file data
    fastr.log.info('Trying to read file with compression {}'.format('ON' if compressed else 'OFF'))
    if compressed:
        try:
            with gzip.open(filename, 'r') as file_handle:
                data = file_handle.read()
        except:
            fastr.log.error('Problem reading gzipped file: {}'.format(filename))
            return False
    else:
        try:
            with open(filename, 'r') as file_handle:
                data = file_handle.read()
        except:
            fastr.log.error('Problem reading normal file: {}'.format(filename))
            return False

    fastr.log.info('Read data from file successfully')

    # Try to read tool doc based on serializer matching the extension
    serializer = ext[1:]
    fastr.log.info('Trying to load file using serializer "{}"'.format(serializer))

    if serializer not in Tool.dumpfuncs:
        fastr.log.error('No matching serializer found for "{}"'.format(serializer))
        return False

    load_func = Serializable.dumpfuncs[serializer].loads

    try:
        doc = load_func(data)
    except Exception as exception:
        fastr.log.error('Could not load data using serializer "{}", encountered exception: {}'.format(serializer,
                                                                                                      exception))
        return False

    return doc


def verify_tool(filename):
    """
    Verify that a file
    """
    # Load the file
    doc = verify_resource_loading(filename)

    if not doc:
        fastr.log.error('Could not load data successfully from  {}'.format(filename))
        return False

    # Match the data to the schema for Tools
    fastr.log.info('Validating data against Tool schema')
    serializer = Tool.get_serializer()

    try:
        doc = serializer.instantiate(doc)
    except Exception as exception:
        fastr.log.error('Encountered a problem when verifying the Tool schema: {}'.format(exception))
        return False

    # Create the Tool object as the final test
    fastr.log.info('Instantiating Tool object')
    try:
        tool = Tool(doc)
        tool.filename = filename
    except Exception as exception:
        fastr.log.error('Encountered a problem when creating the Tool object: {}'.format(exception))
        return False

    fastr.log.info('Loaded tool {} successfully'.format(tool))

    fastr.log.info('Testing tool...')
    try:
        tool.test()
    except fastr.exceptions.FastrValueError as e:
        fastr.log.error('Tool is not valid: {}'.format(e))

    return True


# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fastr
from fastr.core.version import Version
from fastr.datatypes import ValueType


class Boolean(ValueType):
    """
    Datatype representing a boolean
    """

    description = 'A boolean value (True of False)'

    def __str__(self):
        # Format the value properly
        if self.format == 'flag' or self.format is None:
            return {True: '__FASTR_FLAG__TRUE___', False: '__FASTR_FLAG__FALSE__'}[self._value]
        elif self.format == 'yn':
            return {True: 'y', False: 'n'}[self._value]
        elif self.format == 'yes':
            return {True: 'yes', False: 'no'}[self._value]
        elif self.format == 'Yes':
            return {True: 'Yes', False: 'No'}[self._value]
        elif self.format == 'YES':
            return {True: 'YES', False: 'NO'}[self._value]
        elif self.format == 'numeric':
            return {True: '1', False: '0'}[self._value]
        elif self.format == 'string':
            return {True: 'true', False: 'false'}[self._value]
        elif self.format == 'String':
            return {True: 'True', False: 'False'}[self._value]
        elif self.format == 'STRING':
            return {True: 'TRUE', False: 'FALSE'}[self._value]
        elif self.format.startswith('CONST:'):
            if self._value:
                return self.format[6:]
            else:
                return '__FASTR_FLAG__FALSE__'
        elif '|' in self.format:
            options = self.format.split('|', 1)  # In the form of "true|false"
            options = {True: options[0], False: options[1]}
            return options[self._value]
        else:
            fastr.log.warning('Unknown Boolean format ({}), reverting to flag'.format(self.format))
            return {True: '__FASTR_FLAG__TRUE___', False: '__FASTR_FLAG__FALSE__'}[self._value]

    def _validate(self):
        """
        Validate the value of the DataType.

        :return: flag indicating validity of Boolean
        :rtype: bool
        """
        if self.value is None:
            return False
        elif isinstance(self.value, bool):
            return True
        else:
            return False

    @property
    def value(self):
        """
        The value of object instantiation of this DataType.
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Setter function for value property
        """
        translation_table = {0: False, '0': False, 'false': False, '__fastr_flag__false__': False,
                             1: True, '1': False, 'true': True, '__fastr_flag__true___': True}

        if isinstance(value, (str, unicode)):
            value = str(value.lower())

        if isinstance(value, bool):
            self._value = value
        elif value in translation_table:
            self._value = translation_table[value]
        else:
            self._value = None
            fastr.log.debug('Not a valid value for a Boolean ({}), ignoring!'.format(value))





# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nose.tools as nt

from fastr import exceptions
from fastr.core.version import Version


class TestVersion():
    def setup(self):
        # Test different constructor
        self.data1 = Version()
        self.data2 = Version(0, 0, None, 'b', 42)
        self.data3 = Version((1, 2))
        self.data4 = Version([3, 4, [5, 6], 'beta', 3, None, '-'])
        self.data5 = Version('0.1-r3')
        self.data6 = Version('1.2.3.4.5.6.7-beta8_with_suffix')
        self.data7 = Version([3, 4, (5, 6), 'beta', 3, None, '-'])
        self.data8 = Version(u'3.3.2')

    def teardown(self):
        pass

    @nt.with_setup(setup, teardown)
    def test_data(self):
        nt.eq_(self.data1, (0, 0, None, None, None, None, None))
        nt.eq_(self.data2, (0, 0, None, 'b', 42, None, None))
        nt.eq_(self.data3, (1, 2, None, None, None, None, None))
        nt.eq_(self.data4, (3, 4, (5, 6), 'beta', 3, None, '-'))
        nt.eq_(self.data5, (0, 1, None, 're', 3, '', '-'))
        nt.eq_(self.data6, (1, 2, (3, 4, 5, 6, 7), 'beta', 8, '_with_suffix', '-'))
        nt.eq_(self.data7, (3, 4, (5, 6), 'beta', 3, None, '-'))
        nt.eq_(self.data8, (3, 3, (2,), None, None, '', None))

    @nt.with_setup(setup, teardown)
    def test_string(self):
        nt.eq_(str(self.data1), '0.0')
        nt.eq_(str(self.data2), '0.0b42')
        nt.eq_(str(self.data3), '1.2')
        nt.eq_(str(self.data4), '3.4.5.6-beta3')
        nt.eq_(str(self.data5), '0.1-r3')
        nt.eq_(str(self.data6), '1.2.3.4.5.6.7-beta8_with_suffix')
        nt.eq_(str(self.data7), '3.4.5.6-beta3')
        nt.eq_(str(self.data8), '3.3.2')

    @nt.with_setup(setup, teardown)
    def test_major(self):
        nt.eq_(self.data1.major, 0)
        nt.eq_(self.data2.major, 0)
        nt.eq_(self.data3.major, 1)
        nt.eq_(self.data4.major, 3)
        nt.eq_(self.data5.major, 0)
        nt.eq_(self.data6.major, 1)
        nt.eq_(self.data7.major, 3)
        nt.eq_(self.data8.major, 3)

    @nt.with_setup(setup, teardown)
    def test_minor(self):
        nt.eq_(self.data1.minor, 0)
        nt.eq_(self.data2.minor, 0)
        nt.eq_(self.data3.minor, 2)
        nt.eq_(self.data4.minor, 4)
        nt.eq_(self.data5.minor, 1)
        nt.eq_(self.data6.minor, 2)
        nt.eq_(self.data7.minor, 4)
        nt.eq_(self.data8.minor, 3)

    @nt.with_setup(setup, teardown)
    def test_extra(self):
        nt.eq_(self.data1.extra, None)
        nt.eq_(self.data2.extra, None)
        nt.eq_(self.data3.extra, None)
        nt.eq_(self.data4.extra, (5, 6))
        nt.eq_(self.data5.extra, None)
        nt.eq_(self.data6.extra, (3, 4, 5, 6, 7))
        nt.eq_(self.data7.extra, (5, 6))
        nt.eq_(self.data8.extra, (2,))

    @nt.with_setup(setup, teardown)
    def test_extra_string(self):
        nt.eq_(self.data1.extra_string, '')
        nt.eq_(self.data2.extra_string, '')
        nt.eq_(self.data3.extra_string, '')
        nt.eq_(self.data4.extra_string, '.5.6')
        nt.eq_(self.data5.extra_string, '')
        nt.eq_(self.data6.extra_string, '.3.4.5.6.7')
        nt.eq_(self.data7.extra_string, '.5.6')
        nt.eq_(self.data8.extra_string, '.2')

    @nt.raises(exceptions.FastrVersionInvalidError)
    def test_wrong_version_int(self):
        # Cannot use int, should raise an Error
        Version(2)

    @nt.raises(exceptions.FastrVersionInvalidError)
    def test_wrong_version_float(self):
        # Cannot use float, should raise an Error
        Version(3.0)

    @nt.raises(exceptions.FastrVersionInvalidError)
    def test_wrong_version_invalid_str(self):
        Version('test123')

    @nt.raises(exceptions.FastrVersionInvalidError)
    def test_wrong_version_list_too_long(self):
        Version(range(10))




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import urlparse
import fastr
from fastr.data import url
from fastr.datatypes import URLType


class AnalyzeImageFile(URLType):
    description = 'Analyze Image file formate'
    extension = 'hdr'

    @classmethod
    def content(cls, invalue, outvalue=None):
        fastr.log.debug('Determining content of AnalyzeImageFile from invalue "{}" and outvalue "{}"'.format(invalue, outvalue))

        # The suffix translation table
        def substitute_extension(path):
            fastr.log.debug('Substituting: {}'.format(path))
            suffixes = {'.hdr.gz': '.img.gz', '.hdr': '.img'}

            for hdr_suffix, img_suffix in suffixes.items():
                if path.endswith(hdr_suffix):
                    return path[:-len(hdr_suffix)] + img_suffix
            raise ValueError('Extension is not valid for AnalyzeImageFile {}'.format(path))

        if url.isurl(invalue):
            parts = urlparse.urlparse(invalue)

            fastr.log.debug('input URL parts found: {}'.format(parts))
            path = substitute_extension(parts.path)

            in_img = urlparse.urlunparse((parts.scheme, parts.netloc, path, parts.params, parts.query, parts.fragment))
        else:
            in_img = substitute_extension(invalue)
            fastr.log.debug('input path found: {} -> {}'.format(invalue, in_img))

        if outvalue is not None:
            if url.isurl(outvalue):
                parts = urlparse.urlparse(outvalue)

                fastr.log.debug('output URL parts found: {}'.format(parts))
                path = substitute_extension(parts.path)

                out_img = urlparse.urlunparse((parts.scheme, parts.netloc, path, parts.params, parts.query, parts.fragment))
            else:
                out_img = substitute_extension(outvalue)
                fastr.log.debug('output path found: {} -> {}'.format(outvalue, out_img))

            contents = [(invalue, outvalue), (in_img, out_img)]
        else:
            contents = [invalue, in_img]

        return contents



import argparse
import os
import shutil


def main():
    parser = argparse.ArgumentParser(description='Copy file to output directory')
    parser.add_argument('--infile', type=str)
    parser.add_argument('--outdir', type=str)

    args = parser.parse_args()

    print(args)
    print('infile: {}'.format(args.infile))
    print('outdir: {}'.format(args.outdir))

    if not os.path.isdir(args.outdir):
        os.makedirs(args.outdir)

    shutil.copy2(args.infile, args.outdir)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fastr

IS_TEST = True
FAILS = True  # Indicate that this network is supposed to have failing jobs


def create_network():
    network = fastr.Network('failing_network')
    # create sources
    source_1 = network.create_source(fastr.typelist['Int'],
                                     id_='source_1')
    source_2 = network.create_source(fastr.typelist['Int'],
                                     id_='source_2')
    source_3 = network.create_source(fastr.typelist['Int'],
                                     id_='source_3')
    # Create sinks
    sink_1 = network.create_sink(fastr.typelist['Int'], id_='sink_1')
    sink_2 = network.create_sink(fastr.typelist['Int'], id_='sink_2')
    sink_3 = network.create_sink(fastr.typelist['Int'], id_='sink_3')
    sink_4 = network.create_sink(fastr.typelist['Int'], id_='sink_4')
    sink_5 = network.create_sink(fastr.typelist['Int'], id_='sink_5')

    # create nodes
    step_1 = network.create_node('Fail', id_='step_1')
    step_2 = network.create_node('Fail', id_='step_2')
    step_3 = network.create_node('Fail', id_='step_3')

    range_node = network.create_node('Range', id_='range')
    sum_node = network.create_node('Sum', id_='sum')

    # create links
    step_1.inputs['in_1'] = source_1.output
    step_1.inputs['in_2'] = source_2.output
    step_1.inputs['fail_2'] = [False, True, False, True]

    step_2.inputs['in_1'] = source_3.output
    step_2.inputs['in_2'] = source_1.output
    step_2.inputs['fail_1'] = [False, False, True, True]

    step_3.inputs['in_1'] = step_1.outputs['out_1']
    step_3.inputs['in_2'] = step_2.outputs['out_2']

    range_node.inputs['value'] = step_3.outputs['out_1']

    sum_node.inputs['values'] = range_node.outputs['result']

    sink_1.input = step_1.outputs['out_1']
    sink_2.input = step_2.outputs['out_2']
    sink_3.input = step_3.outputs['out_1']
    sink_4.input = range_node.outputs['result']
    sink_5.input = sum_node.outputs['result']

    # Check/Draw/execute network
    return network


def source_data(network):
    fastr.log.info('Creating source data for {}'.format(network.id))
    return {
        'source_1': {'sample_1': 'vfslist://example_data/add_ints/values'},
        'source_2': {'sample_1': 'vfslist://example_data/add_ints/values'},
        'source_3': {'sample_1': 'vfslist://example_data/add_ints/values'},
    }


def sink_data(network):
    fastr.log.info('Creating sink data for {}'.format(network.id))
    return {
        'sink_1': 'vfs://tmp/results/{}/sink_1_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id),
        'sink_2': 'vfs://tmp/results/{}/sink_2_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id),
        'sink_3': 'vfs://tmp/results/{}/sink_3_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id),
        'sink_4': 'vfs://tmp/results/{}/sink_4_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id),
        'sink_5': 'vfs://tmp/results/{}/sink_5_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id),
    }


def main():
    network = create_network()
    network.draw_network(name=network.id, draw_dimension=True)
    network.execute(source_data(network), sink_data(network))


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the VirtualFileSystemValueList plugin for fastr
"""

import urlparse
import fastr
import fastr.exceptions as exceptions
from fastr.core.ioplugin import IOPlugin


class VirtualFileSystemValueList(IOPlugin):
    """
    The VirtualFileSystemValueList an expand-only type of IOPlugin. No URLs
    can actually be fetched, but it can expand a single URL into a larger
    amount of URLs. A ``vfslist://`` URL basically is a url that points to a
    file using vfs. This file then contains a number lines each containing
    another URL.

    If the contents of a file ``vfs://mount/some/path/contents`` would be::

        vfs://mount/some/path/file1.txt
        vfs://mount/some/path/file2.txt
        vfs://mount/some/path/file3.txt
        vfs://mount/some/path/file4.txt

    Then using the URL ``vfslist://mount/some/path/contents`` as source data
    would result in the four files being pulled.

    .. note:: The URLs in a vfslist file do not have to use the ``vfs`` scheme,
              but can use any scheme known to the Fastr system.
    """
    scheme = 'vfslist'

    def __init__(self):
        super(VirtualFileSystemValueList, self).__init__()

    def expand_url(self, url):
        if fastr.data.url.get_url_scheme(url) != 'vfslist':
            raise exceptions.FastrValueError('URL not of vfslist type!')

        # Retrieve base path
        parsed = urlparse.urlparse(url)
        listurl = urlparse.urlunparse(urlparse.ParseResult(scheme='vfs', netloc=parsed.netloc, path=parsed.path, params='', query='', fragment=''))
        listpath = fastr.ioplugins.url_to_path(listurl)

        with open(listpath, 'r') as file_handle:
            data = file_handle.read()

        valuelist = tuple((None, x.strip()) for x in data.strip().split('\n'))
        return valuelist
